"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __exportStar = (this && this.__exportStar) || function(m, exports) {
    for (var p in m) if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.deserialize = exports.serialize = exports.enums = exports.struct = exports.option = exports.map = exports.array = exports.extend = exports.variant = exports.field = void 0;
const types_1 = require("./types");
const writer_1 = require("./writer");
const reader_1 = require("./reader");
const errors_1 = require("./errors");
function addSchemaField(prototype, key, fieldType) {
    let fields = prototype.__schema_fields || [];
    fields.push([key, fieldType]);
    prototype.__schema_fields = fields;
}
function getSchemaFields(prototype) {
    return prototype.__schema_fields || [];
}
function addSchemaVariant(ctor, variant) {
    ctor.prototype.__schema_variant = variant;
    const superClass = Object.getPrototypeOf(ctor.prototype).constructor;
    let enumVariants = superClass.prototype.__schema_enum_variants || [];
    enumVariants.push(ctor);
    superClass.prototype.__schema_enum_variants = enumVariants;
}
function getSchemaVariant(prototype) {
    return prototype.__schema_variant;
}
function isVariantObject(obj) {
    return getSchemaVariant(Object.getPrototypeOf(obj)) !== undefined;
}
function isEnumClass(obj) {
    return getSchemaEnumVariants(obj.prototype) !== undefined;
}
function getSchemaEnumVariants(prototype) {
    return prototype.__schema_enum_variants;
}
function serializeValue(path, value, fieldType, writer) {
    if ((0, types_1.isPrimitiveType)(fieldType)) {
        if (fieldType === 'u8') {
            writer.writeU8(value);
        }
        else if (fieldType === 'u16') {
            writer.writeU16(value);
        }
        else if (fieldType === 'u32') {
            writer.writeU32(value);
        }
        else if (fieldType === 'u64') {
            writer.writeU64(value);
        }
        else if (fieldType === 'usize') {
            writer.writeU64(BigInt(value));
        }
        else if (fieldType === 'bool') {
            writer.writeBool(value);
        }
        else if (fieldType === 'string') {
            writer.writeString(value);
        }
        else if (fieldType === 'u8-array') {
            writer.writeU32(value.length);
            for (let i = 0; i < value.length; i++) {
                writer.writeU8(value[i]);
            }
        }
    }
    else if (typeof fieldType === 'number') {
        if (value.length !== fieldType) {
            (0, errors_1.invalidByteArrayLength)(path, fieldType, value.length);
        }
        writer.writeByteArray(value);
    }
    else {
        const { kind, value: v } = fieldType;
        if (kind === 'option') {
            if (value === undefined || value === null) {
                writer.writeU8(0);
            }
            else {
                writer.writeU8(1);
                serializeValue([...path, '<OptionValue>'], value, v, writer);
            }
        }
        else if (kind === 'array') {
            writer.writeU32(value.length);
            for (let i = 0; i < value.length; i++) {
                serializeValue([...path, `<Array[${i}]>`], value[i], v, writer);
            }
        }
        else if (kind === 'struct') {
            serializeStruct(path, value, writer);
        }
        else if (kind === 'enum') {
            serializeEnum(path, value, writer);
        }
        else if (kind === 'map') {
            writer.writeU32(value.size);
            const [keyType, valueType] = v;
            for (const [k, v] of value) {
                serializeValue([...path, `<Map[key]>`], k, keyType, writer);
                serializeValue([...path, `<Map[value]>`], v, valueType, writer);
            }
        }
        else if (kind === 'extend') {
            if ((0, types_1.hasExtendWriter)(v)) {
                writer.writeExtended(value, v);
            }
            else {
                (0, errors_1.extendedWriterNotFound)(path);
            }
        }
    }
}
function deserializeValue(path, fieldType, reader) {
    if ((0, types_1.isPrimitiveType)(fieldType)) {
        if (fieldType === 'u8') {
            return reader.readU8();
        }
        else if (fieldType === 'u16') {
            return reader.readU16();
        }
        else if (fieldType === 'u32') {
            return reader.readU32();
        }
        else if (fieldType === 'u64') {
            return reader.readU64();
        }
        else if (fieldType === 'bool') {
            return reader.readBool();
        }
        else if (fieldType === 'usize') {
            return Number(reader.readU64());
        }
        else if (fieldType === 'string') {
            return reader.readString();
        }
        else if (fieldType === 'u8-array') {
            let arr = [];
            const length = reader.readU32();
            for (let i = 0; i < length; i++) {
                let n = reader.readU8();
                arr.push(n);
            }
            return Uint8Array.from(arr);
        }
    }
    else if (typeof fieldType === 'number') {
        return reader.readByteArray(fieldType);
    }
    else {
        const { kind, value } = fieldType;
        if (kind === 'array') {
            let arr = [];
            const length = reader.readU32();
            for (let i = 0; i < length; i++) {
                let v = deserializeValue([...path, `<Array[${i}]>`], value, reader);
                arr.push(v);
            }
            return arr;
        }
        else if (kind === 'option') {
            let opt = reader.readBool();
            if (opt) {
                const v = deserializeValue([...path, '<OptionValue>'], value, reader);
                return v;
            }
            else {
                return undefined;
            }
        }
        else if (kind === 'struct') {
            return deserializeStruct(path, value, reader);
        }
        else if (kind === 'enum') {
            return deserializeEnum(path, value, reader);
        }
        else if (kind === 'map') {
            const length = reader.readU32();
            const [keyType, valueType] = value;
            const m = new Map();
            for (let i = 0; i < length; i++) {
                let k = deserializeValue([...path, `<Map[key]>`], keyType, reader);
                let v = deserializeValue([...path, `<Map[value]>`], valueType, reader);
                m.set(k, v);
            }
            return m;
        }
        else if (kind === 'extend') {
            if ((0, types_1.hasExtendReader)(value)) {
                return reader.readExtended(value);
            }
            else {
                (0, errors_1.extendedReaderNotFound)(path);
            }
        }
    }
}
function serializeField(path, obj, field, writer) {
    const [key, fieldType] = field;
    const k = key.toString();
    const value = obj[k];
    serializeValue([...path, k], value, fieldType, writer);
}
function deserializeField(path, obj, field, reader) {
    const [key, fieldType] = field;
    const k = key.toString();
    const value = deserializeValue([...path, k], fieldType, reader);
    obj[k] = value;
    return obj;
}
function serializeEnum(path, obj, writer) {
    const prototype = Object.getPrototypeOf(obj);
    const variant = getSchemaVariant(prototype);
    if (variant !== undefined) {
        writer.writeU8(variant);
        serializeStruct([...path, `<Variant[${variant}]>`], obj, writer);
    }
    else {
        (0, errors_1.invalidEnumField)(path);
    }
}
function serializeStruct(path, obj, writer) {
    const prototype = Object.getPrototypeOf(obj);
    const fields = getSchemaFields(prototype);
    for (const field of fields) {
        serializeField(path, obj, field, writer);
    }
}
function deserializeEnum(path, enumClass, reader) {
    const prototype = enumClass.prototype;
    const enumVariants = getSchemaEnumVariants(prototype);
    if (enumVariants instanceof Array) {
        const i = reader.readU8();
        const variant = enumVariants[i];
        return deserializeStruct([...path, `<Variant[${i}]>`], variant, reader);
    }
    else {
        (0, errors_1.invalidEnumField)(path);
    }
}
function deserializeStruct(path, ctor, reader) {
    const prototype = ctor.prototype;
    const fields = getSchemaFields(prototype);
    let obj = {};
    for (const field of fields) {
        obj = deserializeField(path, obj, field, reader);
    }
    return new ctor(obj);
}
function field(fieldType) {
    return function (target, key) {
        addSchemaField(target.constructor.prototype, key, fieldType);
    };
}
exports.field = field;
function variant(variant) {
    return function (target) {
        addSchemaVariant(target, variant);
    };
}
exports.variant = variant;
function extend(options) {
    return { kind: 'extend', value: options };
}
exports.extend = extend;
function array(elementType) {
    return { kind: 'array', value: elementType };
}
exports.array = array;
function map(keyType, valueType) {
    return { kind: 'map', value: [keyType, valueType] };
}
exports.map = map;
function option(innerType) {
    return { kind: 'option', value: innerType };
}
exports.option = option;
function struct(ctor) {
    return { kind: 'struct', value: ctor };
}
exports.struct = struct;
function enums(enumClass) {
    return { kind: 'enum', value: enumClass };
}
exports.enums = enums;
function serialize(obj) {
    const writer = new writer_1.BinaryWriter();
    if (isVariantObject(obj)) {
        serializeEnum([], obj, writer);
    }
    else {
        serializeStruct([], obj, writer);
    }
    return writer.toArray();
}
exports.serialize = serialize;
function deserialize(classType, data) {
    const reader = new reader_1.BinaryReader(data);
    if (isEnumClass(classType)) {
        return deserializeEnum([], classType, reader);
    }
    else {
        return deserializeStruct([], classType, reader);
    }
}
exports.deserialize = deserialize;
__exportStar(require("./types"), exports);
//# sourceMappingURL=index.js.map